@@ -1,8 +1,12 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 from collections import defaultdict
+import datetime
 
-from trytond.model import Model, ModelView, ModelSQL, fields
+from sql import Null
+
+from trytond.model import Model, ModelView, ModelSQL, fields, Unique, \
+    sequence_ordered
 from trytond.pyson import If, Eval
 from trytond.transaction import Transaction
 from trytond.wizard import Wizard, StateView, StateAction, StateTransition, \
@@ -17,36 +21,46 @@ __all__ = ['Procedure', 'Level', 'Dunning',
 class Procedure(ModelSQL, ModelView):
     'Account Dunning Procedure'
     __name__ = 'account.dunning.procedure'
-    name = fields.Char('Name', required=True, translate=True)
+    name = fields.Char('Name', required=True, translate=True,
+        help="The main identifier of the Dunning Procedure.")
     levels = fields.One2Many('account.dunning.level', 'procedure', 'Levels')
 
 
-class Level(ModelSQL, ModelView):
+class Level(sequence_ordered(), ModelSQL, ModelView):
     'Account Dunning Level'
     __name__ = 'account.dunning.level'
-    _rec_name = 'procedure'
-    sequence = fields.Integer('Sequence')
     procedure = fields.Many2One('account.dunning.procedure', 'Procedure',
         required=True, select=True)
-    days = fields.Integer('Number of Overdue Days')
+    overdue = fields.TimeDelta('Overdue',
+        help="When the level should be applied.")
 
     @classmethod
-    def __setup__(cls):
-        super(Level, cls).__setup__()
-        cls._order.insert(0, ('sequence', 'ASC'))
+    def __register__(cls, module_name):
+        cursor = Transaction().connection.cursor()
+        table = cls.__table_handler__(module_name)
+        sql_table = cls.__table__()
 
-    @staticmethod
-    def order_sequence(tables):
-        table, _ = tables[None]
-        return [table.sequence == None, table.sequence]
+        super(Level, cls).__register__(module_name)
+
+        # Migration from 4.0: change days into timedelta overdue
+        if table.column_exist('days'):
+            cursor.execute(*sql_table.select(sql_table.id, sql_table.days,
+                    where=sql_table.days != Null))
+            for id_, days in cursor.fetchall():
+                overdue = datetime.timedelta(days)
+                cursor.execute(*sql_table.update(
+                        [sql_table.overdue],
+                        [overdue],
+                        where=sql_table.id == id_))
+            table.drop_column('days')
 
     def get_rec_name(self, name):
         return '%s@%s' % (self.procedure.levels.index(self),
             self.procedure.rec_name)
 
     def test(self, line, date):
-        if self.days is not None:
-            return (date - line.maturity_date).days >= self.days
+        if self.overdue is not None:
+            return (date - line.maturity_date) >= self.overdue
 
 _STATES = {
     'readonly': Eval('state') != 'draft',
@@ -58,12 +72,14 @@ class Dunning(ModelSQL, ModelView):
     'Account Dunning'
     __name__ = 'account.dunning'
     company = fields.Many2One('company.company', 'Company', required=True,
+        help="Make the dunning belong to the company.",
         select=True, domain=[
             ('id', If(Eval('context', {}).contains('company'), '=', '!='),
                 Eval('context', {}).get('company', -1)),
             ],
         states=_STATES, depends=_DEPENDS)
     line = fields.Many2One('account.move.line', 'Line', required=True,
+        help="The receivable line to dun for.",
         domain=[
             ('account.kind', '=', 'receivable'),
             ('account.company', '=', Eval('company', -1)),
@@ -80,10 +96,12 @@ class Dunning(ModelSQL, ModelView):
             ('procedure', '=', Eval('procedure', -1)),
             ],
         states=_STATES, depends=_DEPENDS + ['procedure'])
-    blocked = fields.Boolean('Blocked')
+    blocked = fields.Boolean('Blocked',
+        help="Check to block further levels of the procedure.")
     state = fields.Selection([
             ('draft', 'Draft'),
-            ('done', 'Done'),
+            ('waiting', "Waiting"),
+            ('final', "Final"),
             ], 'State', readonly=True)
     active = fields.Function(fields.Boolean('Active'), 'get_active',
         searcher='search_active')
@@ -109,10 +127,23 @@ class Dunning(ModelSQL, ModelView):
     @classmethod
     def __setup__(cls):
         super(Dunning, cls).__setup__()
+        table = cls.__table__()
         cls._sql_constraints = [
-            ('line_unique', 'UNIQUE(line)',
+            ('line_unique', Unique(table, table.line),
                 'Line can be used only once on dunning.'),
             ]
+        cls._active_field = 'active'
+
+    @classmethod
+    def __register__(cls, module):
+        dunning = cls.__table__()
+        super().__register__(module)
+        cursor = Transaction().connection.cursor()
+
+        # Migration from 4.8: rename done state into waiting
+        cursor.execute(*dunning.update(
+                [dunning.state], ['waiting'],
+                where=dunning.state == 'done'))
 
     @staticmethod
     def default_company():
@@ -134,7 +165,7 @@ class Dunning(ModelSQL, ModelView):
 
     @classmethod
     def search_line_field(cls, name, clause):
-        return [('line.' + name,) + tuple(clause[1:])]
+        return [('line.' + clause[0],) + tuple(clause[1:])]
 
     def get_amount(self, name):
         return self.line.debit - self.line.credit
@@ -197,32 +228,43 @@ class Dunning(ModelSQL, ModelView):
 
         set_level = defaultdict(list)
         for dunning in cls.search([
-                    ('state', '=', 'done'),
+                    ('state', '=', 'waiting'),
                     ('blocked', '=', False),
                     ]):
             procedure = dunning.procedure
             levels = procedure.levels
             levels = levels[levels.index(dunning.level) + 1:]
-            for level in levels:
-                if level.test(dunning.line, date):
-                    break
+            if levels:
+                for level in levels:
+                    if level.test(dunning.line, date):
+                        break
+                else:
+                    level = dunning.level
+                if level != dunning.level:
+                    set_level[level].append(dunning)
             else:
-                level = dunning.level
-            if level != dunning.level:
-                set_level[level].append(dunning)
-        for level, dunnings in set_level.iteritems():
-            cls.write(dunnings, {
-                    'level': level.id,
-                    'state': 'draft',
-                    })
+                set_level[None].append(dunning)
+        to_write = []
+        for level, dunnings in set_level.items():
+            if level:
+                to_write.extend((dunnings, {
+                            'level': level.id,
+                            'state': 'draft',
+                            }))
+            else:
+                to_write.extend((dunnings, {
+                            'state': 'final',
+                            }))
+        if to_write:
+            cls.write(*to_write)
 
         lines = MoveLine.search(cls._overdue_line_domain(date))
         dunnings = (cls._get_dunning(line, date) for line in lines)
-        cls.create([d._save_values for d in dunnings if d])
+        cls.save([d for d in dunnings if d])
 
     @classmethod
     def _get_dunning(cls, line, date):
-        procedure = line.party.dunning_procedure
+        procedure = line.dunning_procedure
         if not procedure:
             return
         for level in procedure.levels:
@@ -238,15 +280,17 @@ class Dunning(ModelSQL, ModelView):
 
     @classmethod
     def process(cls, dunnings):
-        cls.write([d for d in dunnings if not d.blocked], {
-                'state': 'done',
+        cls.write([d for d in dunnings
+                if not d.blocked and d.state == 'draft'], {
+                'state': 'waiting',
                 })
 
 
 class CreateDunningStart(ModelView):
     'Create Account Dunning'
     __name__ = 'account.dunning.create.start'
-    date = fields.Date('Date', required=True)
+    date = fields.Date('Date', required=True,
+        help="Create dunning up to this date.")
 
     @staticmethod
     def default_date():
